import unittest

import numpy as np
from ase.quaternions import Quaternion

from muspinsim.utils import (
    Clonable,
    deepmap,
    quat_from_polar,
    zcw_gen,
    eulrange_gen,
    get_xy,
)


class TestUtils(unittest.TestCase):
    def test_clonable(self):
        class TestClass(Clonable):
            def __init__(self):
                self.data = {"x": [1, 2, 3]}

        tc = TestClass()
        # Clone it
        tc2 = tc.clone()
        # Change values
        tc2.data["x"][0] = 2
        # Check it didn't change the original
        self.assertEqual(tc.data["x"][0], 1)

    def test_deepmap(self):

        data = [[1, 2, 3], [4, 5], [6, [7, 8]]]

        def square(x):
            return x**2

        data2 = deepmap(square, data)

        self.assertEqual(data2[0], [1, 4, 9])
        self.assertEqual(data2[1], [16, 25])
        self.assertEqual(data2[2][0], 36)
        self.assertEqual(data2[2][1], [49, 64])

    def test_quat(self):

        q1 = quat_from_polar(np.pi / 4.0, 0)
        z1 = q1.rotate([0, 0, 1])

        self.assertTrue(np.isclose(z1, [2 ** (-0.5), 0, 2 ** (-0.5)]).all())

        theta = 0.6 * np.pi
        phi = 0.4 * np.pi
        ct = np.cos(theta)
        st = np.sin(theta)
        cp = np.cos(phi)
        sp = np.sin(phi)

        q2 = quat_from_polar(theta, phi)
        z2 = q2.rotate([0, 0, 1])

        self.assertTrue(np.isclose(z2, [st * cp, st * sp, ct]).all())

        # Let's check that the inverse works as planned too
        q3 = q2.conjugate()
        z3 = q3.rotate([0, 0, 1])

        self.assertTrue(np.isclose(z3, [-st * cp, st * sp, ct]).all())

    def test_zcw(self):

        N = 1000
        orients = zcw_gen(N)

        # Are these correct? Basic test
        f = 3.0 * np.cos(orients[:, 0]) ** 2 - 1.0
        self.assertAlmostEqual(np.average(f), 0.0, 5)

    def test_eulrange(self):

        N = 10
        ow = eulrange_gen(N)

        A = np.array([[1, 2, 3], [2, 4, 4], [3, 4, -3]])

        Asum = np.zeros((3, 3))
        quats = [Quaternion.from_euler_angles(a, b, c).q for (a, b, c, w) in ow]

        for (a, b, c, w) in ow:
            q = Quaternion.from_euler_angles(a, b, c)
            R = q.rotation_matrix()
            Asum += (R @ A @ R.T) * w

        Asum /= np.sum(ow[:, -1])

        err = np.sum(np.abs(Asum - np.eye(3) * np.trace(A) / 3))

        # Yeah, it's really crude for now, sadly
        self.assertLess(err, 1.0)

    def test_xy(self):

        z = np.array([0, 0, 1])
        x, y = get_xy(z)

        self.assertTrue(np.isclose(x, [1, 0, 0]).all())

        z = np.array([1, 1, 2.0])
        x, y = get_xy(z)
        z /= np.linalg.norm(z)

        self.assertTrue(np.isclose(np.cross(x, y), z).all())

        with self.assertRaises(ValueError):
            get_xy([0, 0, 0])



"""constants.py

Collects methods to retrieve important physical constants
"""

import numpy as np
from scipy import constants as cnst
from soprano.nmr.utils import _get_isotope_data

# Values taken from CODATA on 18/01/2021
# Electron gyromagnetic ratio (MHz/T)
ELEC_GAMMA = -28024.9514242
MU_GAMMA = -(ELEC_GAMMA / 206.7669883)  # Muon gyromagnetic ratio (MHz/T)
MU_TAU = 2.19703  # Muon decay rate (10^-6 s)
# EFG to MHz constant for Quadrupole couplings
# (the total quadrupole coupling in MHz is QCONST*Q*Vzz)
# The factor 1e-37 comes from 1e-6 * 1e-28 * 1e-3
# in order this converts to MHz, from barn to m^2 and the last factor
# accounts for the soprano returning quadrupole moments in millibarn
EFG_2_MHZ = (
    cnst.physical_constants["atomic unit of electric field gradient"][0]
    * cnst.e
    * 1e-37
    / cnst.h
)


def gyromagnetic_ratio(elem="mu", iso=None):
    """Return the gyromagnetic ratio of a given particle

    Return the gyromagnetic ratio of either an atomic nucleus,
    an electron or a muon. The value returned is in MHz/T. It
    corresponds to a frequency, not a pulsation (so it's gamma,
    not gammabar = 2*pi*gamma).

    Keyword Arguments:
        elem {str} -- Element ('e' for electron, 'mu' for muon) (default: {'mu'})
        iso {int} -- Desired isotope. Ignored for 'e' and 'mu'. If not
                     specified, the most naturally abundant isotope is used.
                     (default: {None})
    """

    if elem == "e":
        return ELEC_GAMMA
    elif elem == "mu":
        return MU_GAMMA
    else:
        try:
            val = _get_isotope_data([elem], "gamma", isotope_list=[iso])[0]
        except RuntimeError:
            raise ValueError("Invalid isotope {0} for element {1}".format(iso, elem))

        return val / (2e6 * np.pi)


def quadrupole_moment(elem="mu", iso=None):
    """Return the quadrupole moment of a given particle

    Return the quadrupole moment of either an atomic nucleus,
    an electron or a muon. The value returned is in barn.

    Keyword Arguments:
        elem {str} -- Element ('e' for electron, 'mu' for muon) (default: {'mu'})
        iso {int} -- Desired isotope. Ignored for 'e' and 'mu'. If not
                     specified, the most naturally abundant isotope is used.
                     (default: {None})
    """

    if elem in ("e", "mu"):
        return 0
    else:
        try:
            val = _get_isotope_data([elem], "Q", isotope_list=[iso])[0]
        except RuntimeError:
            raise ValueError("Invalid isotope {0} for element {1}".format(iso, elem))

        return val


def spin(elem="mu", iso=None):
    """Return the spin of a given particle

    Return the intrinsic angular momentum of either an atomic nucleus,
    an electron or a muon. The value returned is in hbar.

    Keyword Arguments:
        elem {str} -- Element ('e' for electron, 'mu' for muon) (default: {'mu'})
        iso {int} -- Desired isotope. Ignored for 'mu'. If used for 'e', this
                     is interpreted as a number of strongly coupled electrons
                     acting as a single spin > 1/2. If not specified, the most
                     naturally abundant isotope is used.
                     (default: {None})
    """

    if elem == "mu":
        return 0.5
    elif elem == "e":
        iso = iso or 1
        if iso < 1 or int(iso) != iso:
            raise ValueError("Invalid multiplicity {0} for electron".format(iso))
        return 0.5 * int(iso)
    else:
        try:
            val = _get_isotope_data([elem], "I", isotope_list=[iso])[0]
        except RuntimeError:
            raise ValueError("Invalid isotope {0} for element {1}".format(iso, elem))

        return val



"""variables.py

Class defining a fitting variable with a starting value and bounds"""

import numpy as np


class FittingVariable(object):
    def __init__(self, name, value=0.0, minv=-np.inf, maxv=np.inf):

        self._name = name
        self._value = float(value)
        self._min = float(minv)
        self._max = float(maxv)

        invalid = ""
        if self._max <= self._min:
            invalid += (
                "Variable {0} has invalid range: "
                "(max value {1} cannot be less than or equal to min value {2}"
                ")\n".format(name, self._max, self._min)
            )
        if self._value > self._max:
            invalid += (
                "Variable {0} has invalid starting value: "
                "(starting value {1} cannot be greater than max value {2})".format(
                    name, self._value, self._max
                )
            )
        if self._value < self._min:
            invalid += (
                "Variable {0} has invalid starting value: "
                "(starting value {1} cannot be less than min value {2})".format(
                    name, self._value, self._min
                )
            )

        if invalid != "":
            raise ValueError(invalid)

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value

    @property
    def bounds(self):
        return (self._min, self._max)
